/* $NetBSD$ */

/*
 * 2009 Jared D. McNeill <jmcneill@invisible.ca>
 * Public domain.
 */

#include <err.h>
#include <signal.h>
#include <unistd.h>

#include <fcntl.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/videoio.h>
#include <dev/wscons/wsconsio.h>

#include "yuv.h"

/* #define DBLBUF */

#ifndef min
#define min(a, b)	((a) < (b) ? (a) : (b))
#endif

static int camerafd, wsdisplayfd;
static bool doquit;

static void
usage(void)
{
	fprintf(stderr, "usage: %s [device]\n", getprogname());
	exit(EXIT_FAILURE);
}

static void
cameractl_restore_display(void)
{
	u_int wsmode;
	int error;

	wsmode = WSDISPLAYIO_MODE_EMUL;
	error = ioctl(wsdisplayfd, WSDISPLAYIO_SMODE, &wsmode);
	if (error)
		warn("WSDISPLAYIO_MODE_EMUL");
}

void
fb_swblit24(uint8_t *fb, struct wsdisplay_fbinfo *fbinfo, u_int stride,
    uint8_t *rgb, u_int rgbw, u_int rgbh)
{
	u_int x = 0, y = 0;
	uint8_t *fbrow, *rgbrow;
	int o = 0, b = fbinfo->depth / 8;

	if (fbinfo->depth == 32)
		o = 1;

	for (y = 0; y < min(fbinfo->height, rgbh); y++) {
		fbrow = fb + (y * stride);
		rgbrow = rgb + (y * rgbw * 3);
		for (x = 0; x < min(fbinfo->width, rgbw); x++) {
			fbrow[x * b + 2 + o] = rgbrow[x * 3 + 0];
			fbrow[x * b + 1 + o] = rgbrow[x * 3 + 1];
			fbrow[x * b + 0 + o] = rgbrow[x * 3 + 2];
		}
	}
}

void
fb_swblit16(uint8_t *fb, struct wsdisplay_fbinfo *fbinfo, u_int stride,
    uint8_t *rgb, u_int rgbw, u_int rgbh)
{
	u_int x = 0, y = 0;
	uint8_t *fbrow, *rgbrow;
	int r, g, b;
	uint16_t pixel;

	for (y = 0; y < min(fbinfo->height, rgbh); y++) {
		fbrow = fb + (y * stride);
		rgbrow = rgb + (y * rgbw * 3);
		for (x = 0; x < min(fbinfo->width, rgbw); x++) {
			r = rgbrow[x * 3 + 0] >> 3;
			g = rgbrow[x * 3 + 1] >> 2;
			b = rgbrow[x * 3 + 2] >> 3;

			pixel = (r << 11) | (g << 5) | b;
			fbrow[x * 2 + 1] = (pixel >> 8);
			fbrow[x * 2 + 0] = (pixel & 0xff);
		}
	}
}

void
fb_swblit(uint8_t *fb, struct wsdisplay_fbinfo *fbinfo, u_int stride,
    uint8_t *rgb, u_int rgbw, u_int rgbh)
{
	switch (fbinfo->depth) {
	case 24:
	case 32:
		fb_swblit24(fb, fbinfo, stride, rgb, rgbw, rgbh);
		return;
	case 16:
		fb_swblit16(fb, fbinfo, stride, rgb, rgbw, rgbh);
		return;
	default:
		fprintf(stderr, "depth %d not supported!\n", fbinfo->depth);
		return;
	}
}

void
sighandler(int sig)
{
	doquit = true;
}

int
main(int argc, const char * const argv[])
{
	struct v4l2_format fmt;
	struct wsdisplay_fbinfo fbinfo;
	const char *videodev = "/dev/video0";
	const char *wsdisplaydev = "/dev/ttyE0";
	uint8_t *buf, *rgb;
	uint8_t *fb;
#ifdef DBLBUF
	uint8_t *fbback;
#endif
	u_int wsmode, stride;
	int error;
	size_t frlen;
	ssize_t rdlen;
	struct sigaction osa;
	static const struct sigaction sa = {
		.sa_handler = sighandler
	};
  
	if (sigaction(SIGHUP, &sa, &osa) == -1)
		err(EXIT_FAILURE, "sigaction(SIGHUP)");
	if (sigaction(SIGINT, &sa, &osa) == -1)
		err(EXIT_FAILURE, "sigaction(SIGINT)");
	if (sigaction(SIGTERM, &sa, &osa) == -1)
		err(EXIT_FAILURE, "sigaction(SIGTERM)");
	if (sigaction(SIGQUIT, &sa, &osa) == -1)
		err(EXIT_FAILURE, "sigaction(SIGQUIT)");

	setprogname(argv[0]);

	if (argc > 2)
		usage();
		/* NOTREACHED */
	if (argc == 2)
		videodev = argv[1];

	atexit(cameractl_restore_display);

	camerafd = open(videodev, O_RDONLY);
	if (camerafd == -1) {
		err(EXIT_FAILURE, "open camera");
	}
	wsdisplayfd = open(wsdisplaydev, O_RDWR);
	if (wsdisplayfd == -1) {
		err(EXIT_FAILURE, "open wsdisplaydev");
	}

	error = ioctl(camerafd, VIDIOC_G_FMT, &fmt);
	if (error) {
		err(EXIT_FAILURE, "VIDIOC_G_FMT");
	}
	printf("video format info\n\theight: %d\n", fmt.fmt.pix.height);
	printf("\twidth: %d\n",fmt.fmt.pix.width);
	printf("\tbytes per line: %d\n", fmt.fmt.pix.bytesperline);
	printf("\timage size: %d\n",fmt.fmt.pix.sizeimage);
	printf("\tpixel format: %04x\n", fmt.fmt.pix.pixelformat);
	printf("\tenum field:%d\n", fmt.fmt.pix.field);

	warnx("%s.%d\n", __func__, __LINE__);

	if (fmt.fmt.pix.pixelformat != V4L2_PIX_FMT_YUYV) {
		err(EXIT_FAILURE, "pixel format not supported\n");
	}

	warnx("%s.%d\n", __func__, __LINE__);
	frlen = fmt.fmt.pix.width * fmt.fmt.pix.bytesperline;
	buf = malloc(frlen);
	if (buf == NULL) {
		err(EXIT_FAILURE, "malloc failed");
	}
	rgb = malloc(fmt.fmt.pix.width * fmt.fmt.pix.height * 3);
	if (rgb == NULL) {
		err(EXIT_FAILURE, "malloc failed");
	}

	warnx("%s.%d\n", __func__, __LINE__);
	wsmode = WSDISPLAYIO_MODE_DUMBFB;
	error = ioctl(wsdisplayfd, WSDISPLAYIO_SMODE, &wsmode);
	warnx("%s.%d\n", __func__, __LINE__);
	if (error == -1) {
		cameractl_restore_display();
		err(EXIT_FAILURE, "WSDISPLAYIO_SMODE");
	}
	warnx("%s.%d\n", __func__, __LINE__);
	error = ioctl(wsdisplayfd, WSDISPLAYIO_GINFO, &fbinfo);
	if (error == -1) {
		err(EXIT_FAILURE, "WSDISPLAYIO_GINFO");
	}
	warnx("%s.%d\n", __func__, __LINE__);
	error = ioctl(wsdisplayfd, WSDISPLAYIO_LINEBYTES, &stride);
	if (error == -1) {
		err(EXIT_FAILURE, "WSDISPLAYIO_LINEBYTES");
	}

	warnx("%s.%d\n", __func__, __LINE__);
	fb = mmap(NULL, fbinfo.height * stride, PROT_READ|PROT_WRITE,
	    MAP_SHARED, wsdisplayfd, 0);
	if (fb == MAP_FAILED) {
		err(EXIT_FAILURE, "mmap");
	}
#ifdef DBLBUF
	fbback = calloc(1, fbinfo.height * stride);
	if (fbback == NULL) {
		err(EXIT_FAILURE, "calloc");
	}
#endif

	warnx("%s.%d\n", __func__, __LINE__);
	memset(fb, 0, fbinfo.height * stride);

	while (!doquit) {
		rdlen = read(camerafd, buf, frlen);
		if (rdlen == -1) {
			err(EXIT_FAILURE, "read camera");
		}
#ifdef DBLBUF
		yuv_decode_rgb(buf, rgb, rdlen, 3);
		fb_swblit(fbback, &fbinfo, stride,
		    rgb, fmt.fmt.pix.width, fmt.fmt.pix.height);
		memcpy(fb, fbback, fbinfo.height * stride);
#else
		if (fbinfo.width == fmt.fmt.pix.width &&
		    fbinfo.height == fmt.fmt.pix.height &&
		    (fbinfo.depth == 24 || fbinfo.depth == 32)) {
			yuv_decode_bgr(buf, fb, rdlen, fbinfo.depth / 8);
		} else {
			yuv_decode_rgb(buf, rgb, rdlen, 3);
			fb_swblit(fb, &fbinfo, stride,
			    rgb, fmt.fmt.pix.width, fmt.fmt.pix.height);
		}
#endif
	}

	return EXIT_SUCCESS;
}

