#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define CLAMP(x, l, h)  (((x) > (h)) ? (h) : (((x) < (l)) ? (l) : (x)))

typedef struct {
	int32_t ug;
	int32_t ub;
	int32_t vg;
	int32_t vr;
} uv_intermediate_t;

/*
 * Converts the U and V components of a YUV pixel to intermediate
 * values used in the YUV to RGB conversion.  These are calculated
 * separately because YUV formats often share the U and V components
 * over several pixels.
 */
static void
uv_to_intermediate(uint8_t u, uint8_t v, uv_intermediate_t *uvi)
{
	/* The author of libccvt asserts that these numbers are from
	   Philips documentation.

	 yy = y << 8;
	 ug =  88 * (u - 128);
	 ub = 454 * (u - 128);
	 vg = 183 * (v - 128);                                                          
	 vr = 359 * (v - 128);                                                          

	 These are calculated from v4l2 documentation, which we use
	 instead: */

#if 1
	 uvi->ug = 100 * (u - 128);
	 uvi->ub = 516 * (u - 128);
	 uvi->vg = 208 * (v - 128);
	 uvi->vr = 409 * (v - 128);
#else
	 uvi->ug =  88 * (u - 128);
	 uvi->ub = 454 * (u - 128);
	 uvi->vg = 183 * (v - 128);
	 uvi->vr = 359 * (v - 128);
#endif
}

/* Takes args as specified below, return r g b between 0 and 255.

   The following YUV -> RGB scheme is taken and used under the GPL
   from the file "ccvt_c1.c" from the library "Colour ConVerT:
   conversion routines from YUV to RGB colour space and back" Version:
   0.3 (24-10-2004 [sic?]) (C) 2001-2002 Nemosoft Unv.
   nemosoft@smcc.demon.nl */
static void
yuv_to_rgb(uint8_t y, uint8_t u, uint8_t v, uv_intermediate_t *uvi, uint8_t rgb[3])
{
	int32_t yy;

	/* y = yuv->y << 8; */
	yy = 298 * (y - 16);	/* v4l documentation */
	rgb[0] = CLAMP((yy + uvi->vr)           >> 8, 0, 255);
	rgb[1] = CLAMP((yy - uvi->ug - uvi->vg) >> 8, 0, 255);
	rgb[2] = CLAMP((yy + uvi->ub)           >> 8, 0, 255);
}

static void
yuv_to_bgr(uint8_t y, uint8_t u, uint8_t v, uv_intermediate_t *uvi, uint8_t rgb[3])
{
	int32_t yy;

	/* y = yuv->y << 8; */
	yy = 298 * (y - 16);	/* v4l documentation */
	rgb[2] = CLAMP((yy + uvi->vr)           >> 8, 0, 255);
	rgb[1] = CLAMP((yy - uvi->ug - uvi->vg) >> 8, 0, 255);
	rgb[0] = CLAMP((yy + uvi->ub)           >> 8, 0, 255);
}

void
yuv_decode_rgb(uint8_t *src, uint8_t *dst, int len, int bpp)
{
	uv_intermediate_t uvi;
	uint8_t *yuyv;

	/* Iterate through each yuyv segment in the buffer
	 * which covers two pixels. */
	for (yuyv = src; yuyv <= (src + len - 4); yuyv += 4) {
		uv_to_intermediate(yuyv[1], yuyv[3], &uvi);
		yuv_to_rgb(yuyv[0], yuyv[1], yuyv[3], &uvi, dst);
		dst += bpp;
		yuv_to_rgb(yuyv[2], yuyv[1], yuyv[3], &uvi, dst);
		dst += bpp;
	}
}

void
yuv_decode_bgr(uint8_t *src, uint8_t *dst, int len, int bpp)
{
	uv_intermediate_t uvi;
	uint8_t *yuyv;

	/* Iterate through each yuyv segment in the buffer
	 * which covers two pixels. */
	for (yuyv = src; yuyv <= (src + len - 4); yuyv += 4) {
		uv_to_intermediate(yuyv[1], yuyv[3], &uvi);
		yuv_to_bgr(yuyv[0], yuyv[1], yuyv[3], &uvi, dst);
		dst += bpp;
		yuv_to_bgr(yuyv[2], yuyv[1], yuyv[3], &uvi, dst);
		dst += bpp;
	}
}
